#include <sys/types.h>
#include <sys/signal.h>
#include <stropts.h>
#include <stdio.h>

#define  SERVER_RENDEZVOUS "/tmp/rendezvous"

char *cmd_name;

main(argc, argv)
int argc;
char *argv[];
{
    int fds[2];                  /* file descriptors to   *
                                  * the pipe we create    */
    struct strrecvfd strrecvfd;  /* contains fd of newly  *
                                  * created pipe          */

    cmd_name = argv[0];

    (void) signal(SIGCHLD, SIG_IGN);


    /*
     * First, create a STREAMS-based pipe.  This is
     * done via the normal pipe system call.
     */

    if (pipe(fds) < 0) {
        fprintf(stderr, "%s: cannot create pipe\n",
                         cmd_name);
        perror("Reason:");
        exit(1);
    }

    /*
     * Now, attach one end of the pipe to a file.
     */

    if (fattach(fds[0], SERVER_RENDEZVOUS) < 0) {
        fprintf(stderr, "%s: cannot attach pipe to file\n",
                         cmd_name);
        perror("Reason:");
        exit(1);
    }

    /*
     *  Push the connld module onto the named end, so
     *  any process that opens the file gets a unique
     *  pipe to us.
     */

    if (ioctl(fds[0], I_PUSH, "connld") < 0) {
        fprintf(stderr, "%s: pushing connld failed\n",
                         cmd_name);
        perror("Reason:");
        exit(1);
    }

    while (ioctl(fds[1], I_RECVFD, &strrecvfd) == 0) {
        switch(fork()) {
          case -1:
              printf("Error: cannot fork!\n");
              perror("Reason:");
              exit(1);
          case 0:
              (void)seteuid(strrecvfd.uid);
              (void)setegid(strrecvfd.gid);
              child(strrecvfd.fd);
              exit(0);
          default:
              /*
               *  Parent simply goes back to
               *  read the next file descriptor
               */
              close(strrecvfd.fd);
              break;
        }
    }
}

int
child(fd)
int fd;
{
    char buf[BUFSIZ];
    FILE *fp;
    int n;

    /*
     *  Read the command from the pipe, execute it
     *  via popen, and send the results back.
     */

    if (read(fd, buf, BUFSIZ) < 0) {
        fprintf(stderr, "%s: cannot read pipe\n", cmd_name);
        perror("Reason:");
        exit(1);
    }

    if ((fp = popen(buf, "r")) == NULL) {
        fprintf(stderr, "%s: popen failed!\n", cmd_name);
        exit(1);
    }

    (void) signal(SIGPIPE, SIG_IGN);
    while ((n = fread(buf, 1, BUFSIZ, fp)) > 0) {
        if (write(fd, buf, n) < 0) {
             fprintf(stderr, "%s: cannot write into pipe\n",
                              cmd_name);
             perror("Reason:");
             exit(1);
        }
    }
    close(fd);
    exit(0);
}
